import 'server-only';
import { db } from '~/server/db';
import * as schema from '~/server/db/schema';
import { eq, and, or, like, not } from 'drizzle-orm';
import { User,
  Relationship,
  UserRelationship,
  RelationshipData,
  Countdown,
  InitialData,
  Message,
  MessageMedia,
  MessageLocation,
  QuickReply,
  QuickReplyOption,
} from '~/server/types';

export const getUser = async (userId: number) => {
  try {
    const users = await db.select().from(schema.users)
      .where(eq(schema.users.id, userId))
    return (users.length > 0) ? users[0] as User : null;
  } catch (error) {
    console.error(error);
    return null;
  }
};

export const getInitialDataByAppleId = async (appleId: string) => {
  try {
    const users = await db.select().from(schema.users)
      .where(eq(schema.users.appleId, appleId))
    if (users.length === 0) return null;
    const user = users[0] as User;
    const userRelationships = await db.select()
      .from(schema.userRelationships)
      .where(eq(schema.userRelationships.userId, user.id))

    let relationshipData: RelationshipData | undefined;
    let countdown: Countdown | undefined;

    if (userRelationships.length > 0) {
      const userRelationship = userRelationships[0] as UserRelationship;
      const relationships = await db.select()
        .from(schema.relationships)
        .where(eq(schema.relationships.id, userRelationship.relationshipId))
      if (relationships.length > 0) {
        const relationship = relationships[0] as Relationship;

        const partners = await db.select()
          .from(schema.users)
          .innerJoin(schema.userRelationships,
            eq(schema.users.id, schema.userRelationships.userId))
          .where(
            and(
              eq(schema.userRelationships.relationshipId, relationship.id),
              not(eq(schema.userRelationships.userId, user.id))
            )
          );
        if (partners.length > 0) {
          const partner = partners[0]?.users as User;
          relationshipData = {
            relationship,
            partner,
          };
          const countdowns = await db.select()
            .from(schema.countdowns)
            .where(eq(schema.countdowns.relationshipId, relationship.id))
            .orderBy(schema.countdowns.date)
            .limit(1);
          if (countdowns.length > 0) {
            countdown = countdowns[0] as Countdown;
          }
        }
      }
    }
    const initialData: InitialData = {
      user,
      relationshipData,
      countdown,
    };
    return initialData;
  } catch (error) {
    console.error(error);
    return null;
  }
};

export const createUser = async (
  appleId: string, email: string,
  fullName: string, pushToken: string
) => {
  try {
    if (!appleId || !email || !fullName || !pushToken) {
      throw new Error("Error: All required fields must be filled");
    }

    // Check if username or email is already taken
    const existingUser = await db.select().from(schema.users)
      .where(or(eq(schema.users.appleId, appleId), eq(schema.users.email, email)));

    if (existingUser.length > 0) {
      throw new Error("Username or email is already in use");
    }

    const newUsers: User[] = await db.insert(schema.users).values({
      appleId, email, fullName, pushToken 
    }).returning() as User[]; // return the newly created user

    if (!newUsers.length || !newUsers[0]?.id)
      throw new Error("Failed to create new user");

    return newUsers[0];

  } catch (error) {
    if (error instanceof Error) {
      throw new Error(`Failed to create new user: ${error.message}`);
    } else {
      throw new Error("Unknown error occurred while creating new user");
    }
  }
};

export const updatePushToken = async (userId: number, pushToken: string): Promise<boolean> => {
  try {
    const result = await db.update(schema.users)
      .set({ pushToken: pushToken })
      .where(
        and(
          eq(schema.users.id, userId),
          not(eq(schema.users.pushToken, pushToken))
        )
      )
      .returning({ updatedId: schema.users.id });

    return result.length > 0;
  } catch (error) {
    console.error('Error updating push token:', error);
    return false;
  }
};

export const updatePfpUrl = async (userId: number, pfpUrl: string) => {
  try {
    const result = await db.update(schema.users)
      .set({ pfpUrl: pfpUrl })
      .where(eq(schema.users.id, userId))
      .returning({ updatedId: schema.users.id });
    return result.length > 0;
  } catch (error) {
    console.error('Error updating pfp url:', error);
  }
};

export const checkRelationshipStatus = async (userId: number): Promise<RelationshipData> => {
  try {
    const user = await getUser(userId);
    if (!user) throw new Error("User not found");
    const userRelationship = await db.select()
      .from(schema.userRelationships)
      .where(eq(schema.userRelationships.userId, user.id))
      .limit(1)
      .then(results => results[0]);
    if (!userRelationship) throw new Error('No relationships found for user');
    const relationship = await db.select()
      .from(schema.relationships)
      .where(eq(schema.relationships.id, userRelationship.relationshipId))
      .limit(1)
      .then(results => results[0] as Relationship);
    if (!relationship) throw new Error('Relationship not found');
    const partner = await db.select()
      .from(schema.users)
      .innerJoin(schema.userRelationships,
        eq(schema.users.id, schema.userRelationships.userId))
      .where(
        and(
          eq(schema.userRelationships.relationshipId, relationship.id),
          not(eq(schema.userRelationships.userId, user.id))
        )
      )
      .limit(1)
      .then(results => results[0]?.users as User);
    if (!partner) throw new Error('No partner found for relationship');
    return { relationship, partner };
  } catch (error) {
    console.error('Error checking relationship status:', error);
    throw error; // Re-throw the error to be handled by the caller
  }
};

export const updateRelationshipStatus = async (
  userId: number, status: 'accepted' | 'rejected'
) => {
  const users = await db.select().from(schema.users)
    .where(eq(schema.users.id, userId));
  const user = users[0] as User;
  if (!user) throw new Error("User not found");
  const userRelationships = await db.select()
    .from(schema.userRelationships)
    .where(eq(schema.userRelationships.userId, user.id));
  if (userRelationships.length === 0) {
    throw new Error('No relationships found for user');
  }
  const userRelationship = userRelationships[0] as UserRelationship;
  const relationships = await db.select()
    .from(schema.relationships)
    .where(eq(schema.relationships.id, userRelationship.relationshipId));
  if (relationships.length === 0) {
    throw new Error('Relationship not found');
  }
  const relationship = relationships[0] as Relationship;
  if (status === 'accepted') {
    await db.update(schema.relationships)
      .set({ isAccepted: true })
      .where(eq(schema.relationships.id, relationship.id));
    const partners = await db.select()
      .from(schema.users)
      .innerJoin(schema.userRelationships,
        eq(schema.users.id, schema.userRelationships.userId))
      .where(
        and(
          eq(schema.userRelationships.relationshipId, relationship.id),
          not(eq(schema.userRelationships.userId, user.id))
        )
      );
    if (partners.length === 0) {
      throw new Error('No partners found for relationship');
    }
    const partner = partners[0]?.users as User;
    const relationshipData: RelationshipData = {
      relationship,
      partner,
    };
    return relationshipData;
  } else if (status === 'rejected') {
    await db.delete(schema.userRelationships)
      .where(eq(schema.userRelationships.id, userRelationship.id));
    await db.delete(schema.relationships)
      .where(eq(schema.relationships.id, relationship.id));
    return null;
  }
};

export const searchUsers = async (userId: number, searchTerm: string) => {
  try {
    const users = await db.select().from(schema.users)
      .where(
        and(
          or(
            like(schema.users.fullName, `%${searchTerm}%`),
            like(schema.users.email, `%${searchTerm}%`)
          ),
          not(eq(schema.users.id, userId))
        )
      );
      if (users.length === 0) throw new Error("No users found");
      return users as User[];
  } catch (error) {
    console.error('Error searching users:', error);
  }
};

export const createRelationshipRequest = async (userId: number, partnerId: number) => {
  try {
    const user = await getUser(userId);
    if (!user) throw new Error("User not found");
    const partner = await getUser(partnerId);
    if (!partner) throw new Error("Partner not found");
    const existingRelationship = await db.select({
      relationshipId: schema.userRelationships.relationshipId,
      status: schema.relationships.isAccepted,
    })
    .from(schema.userRelationships)
    .innerJoin(
      schema.relationships,
      eq(schema.userRelationships.relationshipId, schema.relationships.id)
    )
    .where(
      or(
        eq(schema.userRelationships.userId, user.id),
        eq(schema.userRelationships.userId, partner.id)
      )
    ).limit(1);
    if (existingRelationship.length > 0) {
      throw new Error("Relationship already exists");
    }
    const newRelationship = await db.insert(schema.relationships).values({
      requestorId: user.id,
    }).returning() as Relationship[];
    if (!newRelationship.length || !newRelationship[0]?.id)
      throw new Error("Failed to create new relationship");
    const relationship = newRelationship[0];
    await db.insert(schema.userRelationships).values([
      { userId: userId, relationshipId: relationship.id },
      { userId: partnerId, relationshipId: relationship.id },
    ]);
    return { relationship, partner };
  } catch (error: unknown) {
    if (error instanceof Error) {
      throw new Error(`Failed to create new relationship: ${error.message}`);
    } else {
      throw new Error("Unknown error occurred while creating new relationship");
    }
  }
};

